package tools.descartes.prisma.core;

import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.eclipse.emf.cdo.eresource.CDOResource;
import org.eclipse.emf.cdo.eresource.CDOResourceFolder;
import org.eclipse.emf.cdo.transaction.CDOTransaction;
import org.eclipse.emf.cdo.util.CommitException;
import org.eclipse.emf.cdo.util.ConcurrentAccessException;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.Repository;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import edu.kit.ipd.descartes.mm.applicationlevel.system.SystemFactory;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapeFactory;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfile;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfileFactory;
import tools.descartes.prisma.core.templates.TemplateContext;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class ApplicationModel {
	
	private static final int TRANSACTION_RETRIES = 3;
	
	private final RepositoryConnection connection;
	private final String name;	
	
	public ApplicationModel(RepositoryConnection connection, String name) {
		this.connection = connection;
		this.name = name;
	}
	
	public String getApplicationName() {
		return name;
	}
	
	public String getResourceTypesPath() {
		return "common/defaultresourcetypes.resourcetype";
	}
	
	public String getResourceLandscapePath() {
		return name + "/" + name + ".resourcelandscape";
	}
	
	public String getRepositoryPath() {
		return name + "/" + name + ".repository";
	}
	
	public String getSystemPath() {
		return name + "/" + name + ".system";
	}
	
	public String getDeploymentPath() {
		return name + "/" + name + ".deployment";
	}
	
	public String getUsageProfilePath() {
		return name + "/" + name + ".usageprofile";
	}
	
	public void update(List<ModelTemplate<?,?>> commands) {
		if (commands.isEmpty()) {
			return;
		}
		
		boolean successful = false;
		int retries = TRANSACTION_RETRIES;
		do {
			CDOTransaction transaction = connection.startTransaction();
			TemplateContext context = new TemplateContext(this, transaction);
			for (ModelTemplate<?,?> currentCommand : commands) {
				currentCommand.setContext(context);
			}
			for (ModelTemplate<?,?> currentCommand : commands) {
				currentCommand.apply(null);
			}
			try {
				transaction.commit();
				successful = true;
			} catch (ConcurrentAccessException e) {
				transaction.rollback();
				e.printStackTrace();
			} catch (CommitException e) {
				transaction.rollback();
				e.printStackTrace();
			}
			retries--;
		} while(!successful && retries >= 0);
	}
	
	public void create() {
		try {
			CDOTransaction transaction = connection.startTransaction();
			transaction.getOrCreateResourceFolder(name);
			
			CDOResource resourceLandscapeResource = transaction.getOrCreateResource(getResourceLandscapePath());
			DistributedDataCenter dataCenter;
			if (resourceLandscapeResource.getContents().isEmpty()) {
				dataCenter = ResourcelandscapeFactory.eINSTANCE.createDistributedDataCenter();
				dataCenter.setName(name);
				resourceLandscapeResource.getContents().add(dataCenter);
			} else {
				dataCenter = (DistributedDataCenter)resourceLandscapeResource.getContents().get(0);
			}
			
			CDOResource repositoryResource = transaction.getOrCreateResource(getRepositoryPath());
			if (repositoryResource.getContents().isEmpty()) {
				Repository repository = RepositoryFactory.eINSTANCE.createRepository();
				repository.setName(name);
				repositoryResource.getContents().add(repository);
			}
			
			CDOResource systemResource = transaction.getOrCreateResource(getSystemPath());
			System system;
			if (systemResource.getContents().isEmpty()) {
				system = SystemFactory.eINSTANCE.createSystem();
				system.setName(name);
				systemResource.getContents().add(system);
			} else {
				system = (System)systemResource.getContents().get(0);
			}
			
			CDOResource deploymentResource = transaction.getOrCreateResource(getDeploymentPath());
			if (deploymentResource.getContents().isEmpty()) {
				Deployment deployment = DeploymentFactory.eINSTANCE.createDeployment();
				deployment.setName(name);
				deployment.setSystem(system);
				deployment.setTargetResourceLandscape(dataCenter);
				deploymentResource.getContents().add(deployment);
			}
			
			CDOResource usageProfileResource = transaction.getOrCreateResource(getUsageProfilePath());
			if (usageProfileResource.getContents().isEmpty()) {
				UsageProfile usage = UsageProfileFactory.eINSTANCE.createUsageProfile();
				usage.setName(name);
				usage.setSystem(system);
				usageProfileResource.getContents().add(usage);
			}
			
			transaction.commit();
		} catch(Exception ex) {
			ex.printStackTrace();
		}
	}

}
