package tools.descartes.prisma.parameterization;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import tools.descartes.librede.configuration.ConfigurationFactory;
import tools.descartes.librede.configuration.DataSourceConfiguration;
import tools.descartes.librede.configuration.FileTraceConfiguration;
import tools.descartes.librede.configuration.InputSpecification;
import tools.descartes.librede.configuration.ModelEntity;
import tools.descartes.librede.configuration.Parameter;
import tools.descartes.librede.configuration.Resource;
import tools.descartes.librede.configuration.Service;
import tools.descartes.librede.configuration.TraceFilter;
import tools.descartes.librede.configuration.TraceToEntityMapping;
import tools.descartes.librede.configuration.WorkloadDescription;
import tools.descartes.librede.datasource.csv.CsvDataSource;
import tools.descartes.librede.metrics.Aggregation;
import tools.descartes.librede.metrics.StandardMetrics;
import tools.descartes.librede.units.Ratio;
import tools.descartes.librede.units.RequestCount;
import tools.descartes.librede.units.Time;
import tools.descartes.librede.units.UnitsFactory;

public abstract class InputSpecificationDerivationStrategy {
	
	private Logger log = Logger.getLogger(InputSpecificationDerivationStrategy.class);
	
	private Map<String, FileTraceConfiguration> traces = new HashMap<>();
	
	public abstract InputSpecification derive(WorkloadDescription workload, MappingModel mapping, Path dataLocation);
	
	protected DataSourceConfiguration createWildflyDataSource() {
		DataSourceConfiguration operationsDataSource = ConfigurationFactory.eINSTANCE.createDataSourceConfiguration();
		operationsDataSource.setName("Operations");
		operationsDataSource.setType(CsvDataSource.class.getName());
		Parameter param = ConfigurationFactory.eINSTANCE.createParameter();
		param.setName("TimestampFormat");
		param.setValue("[ms]");
		operationsDataSource.getParameters().add(param);
		return operationsDataSource;
	}
	
	protected DataSourceConfiguration createHypericDataSource() {
		DataSourceConfiguration hypericDataSource = ConfigurationFactory.eINSTANCE.createDataSourceConfiguration();
		hypericDataSource.setName("Hyperic");
		hypericDataSource.setType(CsvDataSource.class.getName());
		
		Parameter param = ConfigurationFactory.eINSTANCE.createParameter();
		param.setName("SkipFirstLine");
		param.setValue("on");
		hypericDataSource.getParameters().add(param);
		param = ConfigurationFactory.eINSTANCE.createParameter();
		param.setName("TimestampFormat");
		param.setValue("[ms]");
		hypericDataSource.getParameters().add(param);
		return hypericDataSource;
	}
	
	protected void addUtilizationTrace(InputSpecification input, DataSourceConfiguration ds, Resource resource, Path dataFile) {
		if (Files.exists(dataFile)) {
			FileTraceConfiguration trace = ConfigurationFactory.eINSTANCE.createFileTraceConfiguration();
			trace.setMetric(StandardMetrics.UTILIZATION);
			trace.setUnit(Ratio.NONE);
			trace.setInterval(UnitsFactory.eINSTANCE.createQuantity(1.0, Time.MINUTES));
			trace.setAggregation(Aggregation.AVERAGE);
			trace.setDataSource(ds);
			trace.setFile(dataFile.toString());
			input.getObservations().add(trace);
					
			TraceToEntityMapping mapping = ConfigurationFactory.eINSTANCE.createTraceToEntityMapping();
			mapping.setEntity(resource);
			mapping.setTraceColumn(2);
			trace.getMappings().add(mapping);
		} else {
			log.warn("File " + dataFile + " does not exist.");
		}
	}	
	
	public List<Service> determineSystemServices(WorkloadDescription workload) {
		List<Service> entryPointServices = new LinkedList<>();
		for (Service curService : workload.getServices()) {
			if (curService.getIncomingCalls().isEmpty()) {
				entryPointServices.add(curService);
			}
		}
		return entryPointServices;
	}	
	
	protected void addResponseTimeTrace(InputSpecification input, DataSourceConfiguration ds, Service service, List<TraceFilter> filters, Path filePath) {
		if (Files.exists(filePath)) {		
			FileTraceConfiguration trace = traces.get("ResponseTime:" + filePath);
			if (trace == null) {
				trace = ConfigurationFactory.eINSTANCE.createFileTraceConfiguration();
				trace.setMetric(StandardMetrics.RESPONSE_TIME);
				trace.setUnit(Time.NANOSECONDS);
				trace.setInterval(UnitsFactory.eINSTANCE.createQuantity(1.0, Time.MINUTES));
				trace.setAggregation(Aggregation.CUMULATIVE_SUM);
				trace.setDataSource(ds);
				trace.setFile(filePath.toString());
				traces.put("ResponseTime:" + filePath, trace);
				input.getObservations().add(trace);
			}
			
			TraceToEntityMapping mapping = ConfigurationFactory.eINSTANCE.createTraceToEntityMapping();
			mapping.setEntity(service);
			mapping.setTraceColumn(10);
	//			mapping.setTraceColumn(10);
			mapping.getFilters().addAll(filters);
			trace.getMappings().add(mapping);
		} else {
			log.warn("File " + filePath + " does not exist.");
		}
	}
	
	protected void addThroughputTrace(InputSpecification input, DataSourceConfiguration ds, ModelEntity entity, List<TraceFilter> filters, Path filePath) {
		if (Files.exists(filePath)) {
			FileTraceConfiguration trace = traces.get("Throughput:" + filePath);
			if (trace == null) {
				trace = ConfigurationFactory.eINSTANCE.createFileTraceConfiguration();
				trace.setMetric(StandardMetrics.DEPARTURES);
				trace.setUnit(RequestCount.REQUESTS);
				trace.setInterval(UnitsFactory.eINSTANCE.createQuantity(1.0, Time.MINUTES));
				trace.setAggregation(Aggregation.CUMULATIVE_SUM);
				trace.setDataSource(ds);
				trace.setFile(filePath.toString());
				traces.put("Throughput:" + filePath, trace);
				input.getObservations().add(trace);
			}
	
			TraceToEntityMapping mapping = ConfigurationFactory.eINSTANCE.createTraceToEntityMapping();
			mapping.setEntity(entity);
			mapping.setTraceColumn(9);
	//		mapping.setTraceColumn(8);
			mapping.getFilters().addAll(filters);
			trace.getMappings().add(mapping);
		} else {
			log.warn("File " + filePath + " does not exist.");
		}
	}
	
	
	private List<TraceFilter> createSignatureFilter(AssemblyContext assembly, Signature signature, int componentColumn, int interfaceColumn, int operationColumn) {
		// The signature name currently consists of the interface name and the operation name separated by a #.
		// In the file traces, only the mere operation names are saved.
		String[] name = signature.getName().split("#");
		String operation = name[name.length - 1];
		
		List<TraceFilter> filters = new LinkedList<TraceFilter>();
		filters.add(createTraceFilter(componentColumn, assembly.getName()));
		filters.add(createTraceFilter(interfaceColumn, signature.getInterface().getName()));
		filters.add(createTraceFilter(operationColumn, operation));
		return filters;
	}
	
	private TraceFilter createTraceFilter(int column, String value) {
		TraceFilter filter = ConfigurationFactory.eINSTANCE.createTraceFilter();
		filter.setTraceColumn(column);
		filter.setValue(value);
		return filter;
	}
	
	protected List<TraceFilter> createCallFilter(AssemblyContext assembly, Signature signature) {
		List<TraceFilter> filters = createSignatureFilter(assembly, signature, 0, 1, 2);
		filters.add(createTraceFilter(4, ""));
		return filters;
	}
	
	protected List<TraceFilter> createCallFilter(AssemblyContext assembly, Signature signature, AssemblyContext calledAssembly, Signature calledSignature) {
		List<TraceFilter> filters = createSignatureFilter(assembly, signature, 0, 1, 2);
		filters.addAll(createSignatureFilter(calledAssembly, calledSignature, 4, 5, 6));
		return filters;
	}	
}
